@@ -1,32 +1,32 @@
-
 use super::*;
+use approx::assert_relative_eq;
 
 #[test]
 fn test_var() {
     let mut s = SymbolTable::new();
     let a_id = s.add_variable("a", 0.).unwrap().unwrap();
     let mut e = Expression::new("a + a / 2", s).unwrap();
-    assert_eq!(e.value(), 0.);
-    e.symbols().set_value(a_id, 2.);
-    assert_eq!(e.value(), 3.);
+    assert_relative_eq!(e.value(), 0.);
+    e.symbols().value_cell(a_id).set(2.);
+    assert_relative_eq!(e.value(), 3.);
 }
 
 #[test]
 fn test_constant() {
     let mut s = SymbolTable::new();
     s.add_constant("a", 2.).unwrap();
-    let e = Expression::new("a + a / 2", s).unwrap();
-    assert_eq!(e.value(), 3.);
+    let mut e = Expression::new("a + a / 2", s).unwrap();
+    assert_relative_eq!(e.value(), 3.);
 }
 
 #[test]
 fn test_string() {
     let mut s = SymbolTable::new();
-    let s_id = s.add_stringvar("s", b"string").unwrap().unwrap();
+    let s_id = s.add_stringvar("s", "string").unwrap().unwrap();
     let mut e = Expression::new("s[] + 1", s).unwrap();
-    assert_eq!(e.value(), 7.);
-    e.symbols().set_string(s_id, b"string2");
-    assert_eq!(e.value(), 8.);
+    assert_relative_eq!(e.value(), 7.);
+    e.symbols_mut().set_string(s_id, "string2");
+    assert_relative_eq!(e.value(), 8.);
 }
 
 #[test]
@@ -34,9 +34,9 @@ fn test_vector() {
     let mut s = SymbolTable::new();
     let vec_id = s.add_vector("v", &[0., 1., 2., 3.]).unwrap().unwrap();
     let mut e = Expression::new("v[] + v[1] + 1", s).unwrap();
-    assert_eq!(e.value(), 6.);
-    e.symbols().mut_vector(vec_id).unwrap()[1] = 2.;
-    assert_eq!(e.value(), 7.);
+    assert_relative_eq!(e.value(), 6.);
+    e.symbols_mut().vector_mut(vec_id)[1] = 2.;
+    assert_relative_eq!(e.value(), 7.);
 }
 
 #[test]
@@ -54,16 +54,17 @@ fn test_vector_out_of_bounds() {
 fn test_funcs() {
     let mut s = SymbolTable::new();
     s.add_func1("add_one", |a| a + 1.).unwrap();
-    let e = Expression::new("add_one(1)", s).unwrap();
-    assert_eq!(e.value(), 2.);
+    let mut e = Expression::new("add_one(1)", s).unwrap();
+    assert_relative_eq!(e.value(), 2.);
 
     let mut s = SymbolTable::new();
     s.add_func1("one", |a| a).unwrap();
     s.add_func2("two", |a, b| a + b).unwrap();
     s.add_func3("three", |a, b, c| a + b + c).unwrap();
     s.add_func4("four", |a, b, c, d| a + b + c + d).unwrap();
-    let e = Expression::new("one(1) + two(1, 1) + three(1, 1 ,1) + four(1, 1, 1, 1)", s).unwrap();
-    assert_eq!(e.value(), 10.);
+    let mut e =
+        Expression::new("one(1) + two(1, 1) + three(1, 1 ,1) + four(1, 1, 1, 1)", s).unwrap();
+    assert_relative_eq!(e.value(), 10.);
 }
 
 #[test]
@@ -88,20 +89,44 @@ fn test_parse_err() {
 fn test_resolver() {
     let mut s = SymbolTable::new();
     s.add_variable("a", 1.).unwrap().unwrap();
-    let (mut expr, vars) = Expression::with_vars("a + 1 + b", s).unwrap();
-    assert_eq!(vars, [("b".to_string(), 1)]);
-    assert_eq!(expr.value(), 2.);
-    expr.symbols().set_value(1, 1.);
-    assert_eq!(expr.value(), 3.);
+    let mut expr = Expression::handle_unknown("a + b + c + s[] + v[]", s, |name, s| {
+        match name {
+            "b" => {
+                s.add_variable(name, 1.).unwrap();
+            }
+            "c" => {
+                s.add_constant(name, 1.).unwrap();
+            }
+            "s" => {
+                s.add_stringvar(name, "string").unwrap();
+            }
+            "v" => {
+                s.add_vector(name, &[1., 2., 3.]).unwrap();
+            }
+            _ => {}
+        }
+        Ok(())
+    })
+    .unwrap();
+    assert_relative_eq!(expr.value(), 12.);
+}
+
+#[test]
+fn test_auto_resolver() {
+    let (mut expr, vars) = Expression::parse_vars("a + b", SymbolTable::new()).unwrap();
+    assert_eq!(vars, vec![("a".to_string(), 0), ("b".to_string(), 1)]);
+    assert_relative_eq!(expr.value(), 0.);
+    expr.symbols().value_cell(0).set(1.);
+    assert_relative_eq!(expr.value(), 1.);
 }
 
 #[test]
 fn test_names() {
     let mut s = SymbolTable::new();
     s.add_variable("a", 1.).unwrap().unwrap();
-    s.add_stringvar("s", b"value").unwrap().unwrap();
+    s.add_stringvar("s", "value").unwrap().unwrap();
     s.add_vector("v", &[1., 2.]).unwrap().unwrap();
-    let (mut expr, _) = Expression::with_vars("a + 1 + b + s[] + v[]", s).unwrap();
+    let (expr, _) = Expression::parse_vars("a + 1 + b + s[] + v[]", s).unwrap();
     assert_eq!(expr.symbols().get_variable_names(), vec!["a", "b"]);
     assert_eq!(expr.symbols().get_stringvar_names(), vec!["s"]);
     assert_eq!(expr.symbols().get_vector_names(), vec!["v"]);
@@ -111,7 +136,7 @@ fn test_names() {
 fn test_clear() {
     let mut s = SymbolTable::new();
     s.add_variable("a", 1.).unwrap().unwrap();
-    s.add_stringvar("s", b"value").unwrap().unwrap();
+    s.add_stringvar("s", "value").unwrap().unwrap();
     s.add_vector("v", &[1., 2.]).unwrap().unwrap();
     s.clear_variables();
     s.clear_strings();
@@ -125,35 +150,34 @@ fn test_clear() {
 fn test_const() {
     let mut s = SymbolTable::new();
     s.add_constant("a", 1.).unwrap();
-    let expr = Expression::new("a + 1", s).unwrap();
-    assert_eq!(expr.value(), 2.);
+    let mut expr = Expression::new("a + 1", s).unwrap();
+    assert_relative_eq!(expr.value(), 2.);
 }
 
-
 #[test]
 fn test_ids() {
     let mut s = SymbolTable::new();
     let a_id = s.add_variable("a", 1.).unwrap().unwrap();
-    let b_id = s.add_variable("b", 1.).unwrap().unwrap();
-    let s_id = s.add_stringvar("s", b"value").unwrap().unwrap();
+    let s_id = s.add_stringvar("s", "value").unwrap().unwrap();
     let v_id = s.add_vector("v", &[1., 2.]).unwrap().unwrap();
-    assert_eq!(s.get_var_id("a"), Some(a_id));
-    assert_eq!(s.get_var_id("b"), Some(b_id));
-    assert_eq!(s.get_var_id("c"), None);
-    assert_eq!(s.get_string_id("s"), Some(s_id));
-    assert_eq!(s.get_vec_id("v"), Some(v_id));
+    assert_eq!(s.get_var_id("a").unwrap(), Some(a_id));
+    assert_eq!(s.get_var_id("c").unwrap(), None);
+    assert_eq!(s.get_string_id("s").unwrap(), Some(s_id));
+    assert_eq!(s.get_vec_id("v").unwrap(), Some(v_id));
+    assert_eq!(s.get_vec_id("s").unwrap(), None);
 }
 
 #[test]
 fn test_clone() {
     let mut s = SymbolTable::new();
     s.add_variable("a", 1.).unwrap().unwrap();
-    s.add_stringvar("s", b"s").unwrap().unwrap();
+    s.add_stringvar("s", "s").unwrap().unwrap();
     s.add_vector("v", &[1., 2.]).unwrap().unwrap();
     s.add_func1("func", |x| x + 1.).unwrap();
-    let expr = Expression::new("a + s[] + v[0] + func(0)", s).unwrap();
-    assert_eq!(expr.value(), 4.);
-    assert_eq!(expr.clone().value(), 4.);
+    let mut expr = Expression::new("a + s[] + v[0] + func(0)", s).unwrap();
+    assert_relative_eq!(expr.value(), 4.);
+    assert_relative_eq!(expr.clone().value(), 4.);
+    assert_eq!(format!("{:?}", expr), format!("{:?}", expr.clone()));
 }
 
 #[test]
@@ -161,16 +185,16 @@ fn test_send() {
     use std::thread;
     let mut s = SymbolTable::new();
     let a_id = s.add_variable("a", 1.).unwrap().unwrap();
-    let s_id = s.add_stringvar("s", b"s").unwrap().unwrap();
+    let s_id = s.add_stringvar("s", "s").unwrap().unwrap();
     let v_id = s.add_vector("v", &[1.]).unwrap().unwrap();
     let mut e = Expression::new("a + s[] + v[0]", s).unwrap();
-    assert_eq!(e.value(), 3.);
+    assert_relative_eq!(e.value(), 3.);
 
     thread::spawn(move || {
-        assert_eq!(e.value(), 3.);
-        e.symbols().set_value(a_id, 2.);
-        e.symbols().set_string(s_id, b"s2");
-        e.symbols().mut_vector(v_id).unwrap()[0] = 2.;
-        assert_eq!(e.value(), 6.);
+        assert_relative_eq!(e.value(), 3.);
+        e.symbols().value_cell(a_id).set(2.);
+        e.symbols_mut().set_string(s_id, "s2");
+        e.symbols_mut().vector_mut(v_id)[0] = 2.;
+        assert_relative_eq!(e.value(), 6.);
     });
 }
